/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    testSegImage.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**  Regression tests for segmentation images.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>

#include <cxcore.h>
#include <cv.h>
#include <highgui.h>

#include "svlBase.h"
#include "svlVision.h"

#define WINDOW_NAME "testSegImage"

using namespace std;

// globals ------------------------------------------------------------------

int gScale = 1;
svlSegImage *gSegImage = NULL;
IplImage *gCanvas = NULL;
IplImage *gView = NULL;

vector<pair<int, int> > gAdjList;
vector<vector<CvPoint> > gEdgels;
vector<pair<CvPoint, set<int> > > gVertices;

// prototypes ---------------------------------------------------------------

void mouseCallback(int event, int x, int y, int flags, void *params);
void drawPlus(IplImage *image, int x, int y, unsigned char red,
    unsigned char green, unsigned char blue);

// main ---------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << "\n";
    cerr << "USAGE: ./testSegImage [OPTIONS] <image> <segmentation>\n";
    cerr << "OPTIONS:\n" 
         << "  -scale <n>        :: rescale image\n"
         << SVL_STANDARD_OPTIONS_USAGE 
         << endl;
}

int main(int argc, char *argv[])
{
    // read commandline parameters
    const int NUM_REQUIRED_PARAMETERS = 2;

    char **args = argv + 1;
    while (--argc > NUM_REQUIRED_PARAMETERS) {
        SVL_PROCESS_STANDARD_OPTIONS(args, argc)
        else if (!strcmp(*args, "-scale")) {
            gScale = atoi(*(++args)); argc -= 1;
            SVL_ASSERT(gScale > 0);
        } else {
            SVL_LOG(SVL_LOG_FATAL, "unrecognized option " << *args);
        }
        args++;
    }

    svlCodeProfiler::tic(svlCodeProfiler::getHandle("main"));
    if (argc != NUM_REQUIRED_PARAMETERS) {
        usage();
        return -1;
    }

    // load segmentation image
    gSegImage = new svlSegImage(args[0], args[1]);
    assert(gSegImage != NULL);
    SVL_LOG(SVL_LOG_VERBOSE, gSegImage->numSegments() << " superpixels");
    gAdjList = gSegImage->getAdjacencyList();
    SVL_LOG(SVL_LOG_VERBOSE, gAdjList.size() << " neighbors");
    gEdgels = gSegImage->getEdgels();
    SVL_LOG(SVL_LOG_VERBOSE, gEdgels.size() << " edgels");
    gVertices = gSegImage->getVertices();
    SVL_LOG(SVL_LOG_VERBOSE, gVertices.size() << " vertices");

    gCanvas = cvCloneImage(gSegImage->visualize());
    gView = cvCreateImage(cvSize(gScale * gCanvas->width, gScale * gCanvas->height),
        gCanvas->depth, gCanvas->nChannels);
    cvResize(gCanvas, gView, CV_INTER_NN);

    cvNamedWindow(WINDOW_NAME, CV_WINDOW_AUTOSIZE);
    cvShowImage(WINDOW_NAME, gView);
    cvSetMouseCallback(WINDOW_NAME, &mouseCallback, NULL);
        
    // enter event processing loop
    int ch = -1;
    while ((ch != 'q') && (ch != 27)) {
        ch = cvWaitKey(100) & 0x00ff;
    }

    // free memory
    cvReleaseImage(&gView);
    cvReleaseImage(&gCanvas);
    delete gSegImage;
    cvDestroyAllWindows();
    svlCodeProfiler::toc(svlCodeProfiler::getHandle("main"));
    svlCodeProfiler::print(cerr);
    return 0;
}

// mosue callback 
void mouseCallback(int event, int x, int y, int flags, void *params)
{
    switch (event) {
    case CV_EVENT_LBUTTONDOWN:
        // todo
        break;
    case CV_EVENT_MOUSEMOVE:
        int segId = gSegImage->getSegment(cvPoint(x / gScale, y / gScale));
        cvCopyImage(gSegImage->visualize(), gCanvas);
        
        gSegImage->colorSegment(gCanvas, segId, 0x7f, 0x00, 0x00, 1.0);

        for (int i = 0; i < (int)gAdjList.size(); i++) {
            if ((gAdjList[i].first != segId) &&
                (gAdjList[i].second != segId))
                continue;

            for (int j = 0; j < (int)gEdgels[i].size(); j++) {
		int baseIndx = gEdgels[i][j].y * gCanvas->widthStep + 3 * gEdgels[i][j].x;
		gCanvas->imageData[baseIndx + 1] = 0xff;
            }
        }

        for (int i = 0; i < (int)gVertices.size(); i++) {
            drawPlus(gCanvas, gVertices[i].first.x, gVertices[i].first.y,
                0x00, 0x00, 0xff);
        }

        for (int i = 0; i < (int)gEdgels.size(); i++) {
            CvPoint c = gEdgels[i][gEdgels[i].size() / 2];
            drawPlus(gCanvas, c.x, c.y, 0x00, 0xff, 0x00);
        }

        for (int i = 0; i < gSegImage->numSegments(); i++) {
            CvPoint c = gSegImage->getCentroid(i);
            drawPlus(gCanvas, c.x, c.y, 0xff, 0x00, 0x00);
        }

        if (gScale == 1) {
            cvShowImage(WINDOW_NAME, gCanvas);
        } else {
            cvResize(gCanvas, gView, CV_INTER_NN);
            cvShowImage(WINDOW_NAME, gView);
        }
        break;
    }
}

void drawPlus(IplImage *image, int x, int y, unsigned char red,
    unsigned char green, unsigned char blue)
{
    int baseIndx = y * image->widthStep + 3 * x;
    image->imageData[baseIndx + 0] = blue;
    image->imageData[baseIndx + 1] = green;
    image->imageData[baseIndx + 2] = red;
    
    if (x < image->width - 1) {
        image->imageData[baseIndx + 3] = blue;
        image->imageData[baseIndx + 4] = green;
        image->imageData[baseIndx + 5] = red;
    }
    if (x > 0) {
        image->imageData[baseIndx - 3] = blue;
        image->imageData[baseIndx - 2] = green;
        image->imageData[baseIndx - 1] = red;
    }
    if (y < image->height - 1) {
        image->imageData[baseIndx + image->widthStep] = blue;
        image->imageData[baseIndx + image->widthStep + 1] = green;
        image->imageData[baseIndx + image->widthStep + 2] = red;
    }
    if (y > 0) {
        image->imageData[baseIndx - image->widthStep] = blue;
        image->imageData[baseIndx - image->widthStep + 1] = green;
        image->imageData[baseIndx - image->widthStep + 2] = red;
    }
}
