/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    filterPatchDictionary.cpp
** AUTHOR(S):   Olga Russakovsky <olga@cs.stanford.edu>
**                based on trainObjectDetector.cpp by Steve Gould
** DESCRIPTION:
**   Trims the dictionary down to a more manageable size; eliminates
**   features that are not very informative on the training set
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <sys/types.h>
#include <vector>

#if defined(_WIN32)||defined(WIN32)||defined(__WIN32__)
#include "win32/dirent.h"
#else
#include <dirent.h>
#endif

#include "cv.h"
#include "cxcore.h"
#include "highgui.h"

#include "svlBase.h"
#include "svlML.h"
#include "svlVision.h"

using namespace std;

// Prototypes -----------------------------------------------------------------

vector<vector<double> > readFromCache(const char* dirName, int maxImages, bool binary);

// Main -----------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << endl;
    cerr << "USAGE: ./filterPatchDictionary <dictionary> <+ cache dir> <- cache dir>" << endl;
    cerr << "OPTIONS:" << endl
	 << "  -o <filename>       :: output filename for the new dictionary" << endl
	 << "  -cc <threshold>     :: threshold for cross-correlation of patches (default: 0.9)" << endl
	 << "  -F <threshold>      :: threshold for F-score of patches (default: 0.75)" << endl
	 << "  -minSize <n>        :: minimum number of features expected (overrides -F, not -cc)" << endl
	 << "  -maxInstances <n>   :: maximum training images for each class (default: 1000)" << endl
	 << "  -binary             :: the cache dirs are really names of binary files" << endl
         << "  -v                  :: verbose" << endl
	 << endl;
}

bool patch_scores_sort_decreasing(pair<float, int> i, pair<float, int> j) {
  return i.first > j.first;
}

int main(int argc, char *argv[])
{
    // read commandline parameters
    const int NUM_REQUIRED_PARAMETERS = 3;

    const char *posCacheDir = NULL;
    const char *negCacheDir = NULL;

    const char *dictionaryFilename = NULL;
    const char *outputFilename = NULL;
    int maxTrainingInstances = 1000;
    double CCthreshold = 0.9;
    double Fthreshold = 0.75;
    int minNumPatches = 0;

    bool binary = false;

    char **args = argv + 1;
    while (--argc > NUM_REQUIRED_PARAMETERS) {
      if (!strcmp(*args, "-maxInstances")) {
	maxTrainingInstances = atoi(*(++args)); argc--;
      } else if (!strcmp(*args, "-o")) {
	outputFilename = *(++args);          argc--;
      } else if (!strcmp(*args, "-cc")) {
	CCthreshold = atof(*(++args));       argc--;
      } else if (!strcmp(*args, "-F")) {
	Fthreshold = atof(*(++args));         argc--;
      } else if (!strcmp(*args, "-minSize")) {
	minNumPatches = atoi(*(++args));      argc--;
      } else if (!strcmp(*args, "-binary")) {
	binary = true;
      } else if (!strcmp(*args, "-v")) {
	svlLogger::setLogLevel(SVL_LOG_VERBOSE);
      } else {
	SVL_LOG(SVL_LOG_FATAL, "unrecognized option " << *args);
      }
      args++;
    }

    if (argc != NUM_REQUIRED_PARAMETERS) {
        usage();
        return -1;
    }

    dictionaryFilename = args[0];
    posCacheDir = args[1];
    negCacheDir = args[2];

    // read the dictionary
    svlPatchDictionary dict;
    dict.read(dictionaryFilename);
    
    // load positive and negative training samples from cache
    SVL_LOG(SVL_LOG_VERBOSE, "Reading positive samples...");
    vector<vector<double> > posSamples = readFromCache(posCacheDir, maxTrainingInstances, binary);
    SVL_LOG(SVL_LOG_VERBOSE, "..." << posSamples.size() << " samples read");
    SVL_LOG(SVL_LOG_VERBOSE, "Reading negative samples...");
    vector<vector<double> > negSamples = readFromCache(negCacheDir, maxTrainingInstances, binary);
    SVL_LOG(SVL_LOG_VERBOSE, "..." << negSamples.size() << " samples read");

    unsigned numPatches = dict.numFeatures();

    // make sure all samples were correctly cached
    for (int j = (int)posSamples.size() - 1; j >= 0; j--) {
      if (posSamples[j].size() != numPatches) {
	posSamples.erase(posSamples.begin() + j);
	SVL_LOG(SVL_LOG_WARNING, "Had to erase posSample " << j << ", incorrect num features");
      }
    }

    for (int j = (int)negSamples.size() - 1; j >= 0; j--) {
      if (negSamples[j].size() != numPatches) {
	negSamples.erase(negSamples.begin() + j);
	SVL_LOG(SVL_LOG_WARNING, "Had to erase negSample " << j << ", incorrect num features");
      }
    }

    vector<pair<float, int> > patchFscores(numPatches);

    // analyze each of the dictionary features in turn
    // to see how good it is at discriminating between +/- examples
    // (note: the +/- examples are weighted s.t. each set contributed 1/2 the weight
    // even though the number of each might be different)
    for (unsigned i = 0; i < numPatches; i++) {

      svlClassifierResponseList c;
      for (unsigned j = 0; j < posSamples.size(); j++)
	insertScore(c, posSamples[j][i], true);
      for (unsigned j = 0; j < negSamples.size(); j++)
	insertScore(c, negSamples[j][i], false);

      normalizeScores(c);
      svlPRcurve p;
      createPRcurve(c, p);
      patchFscores[i] = make_pair(getHighestFscore(p), i);
    }

    // sort the patches in order of Fscore, highest first
    stable_sort(patchFscores.begin(), patchFscores.end(), patch_scores_sort_decreasing);

    // these are indexed by the current ordering of patches
    vector<bool> availableForSelection(numPatches, true);

    // these are indexed by the original ordering, corresponding to the 
    // dictionary entries
    vector<bool> selected(numPatches, false);
    int numSelected = 0;

    // iterate over patches in order of decreasing Fscores
    for (unsigned i = 0; (i < numPatches) && (patchFscores[i].first > Fthreshold || numSelected < minNumPatches); i++) {
      if (availableForSelection[i]) {
	SVL_LOG(SVL_LOG_VERBOSE, "Analyzing patch " << i << " of " << numPatches
		<< ", with value " << patchFscores[i].first);
	int index1 = patchFscores[i].second;
	selected[index1] = true;
	numSelected++;

	int num = 0;
	for (unsigned j = i+1; j < numPatches; j++) {
	  if (availableForSelection[j]) {
	    int index2 = patchFscores[j].second;
	    if (dict.entriesTooCorrelated(index1, index2, CCthreshold)) {
		availableForSelection[j] = false;
		num++;
	    }
	  } 
	}
	SVL_LOG(SVL_LOG_VERBOSE, "Deleted " << num << " patches as a result");
      }
    }

    dict.filterEntries(selected);

    SVL_LOG(SVL_LOG_MESSAGE, "Trimmed dictionary to " << dict.numFeatures() << " patches");

    dict.write(outputFilename);

    return 0;
}












// Helper functions ----------------------------------------------------------
// (shared with trainObjectDetector; really should decompose out)

vector<vector<double> > readFromCache(const char *dirName, int maxImages,
				      bool binary)
{
  vector<vector<double> > fv;
  
  if (!binary) {
    vector<string> filenames = svlDirectoryListing(dirName, ".txt");

    for (unsigned i = 0; i < filenames.size() && i < (unsigned)maxImages; i++) {
      ifstream ifs(filenames[i].c_str());
      if (ifs.fail()) {
	SVL_LOG(SVL_LOG_WARNING, "...failed on " << filenames[i]);
      } else {
	vector<double> v;
	double d;
	while (1) {
	  ifs >> d;
	  if (ifs.fail()) break;
	  v.push_back(d);
	}
	fv.push_back(v);
      }
      SVL_LOG(SVL_LOG_VERBOSE, "reading " << filenames[i]);
    }
  } else {
    ifstream ifs(dirName, std::ios::in|std::ios::binary);

    while (fv.size() < (unsigned)maxImages) {
      int size = -1;
      ifs.read((char *) &size, sizeof(int));
      if (ifs.fail() || size < 0) break;

      vector<double> v(size);
      for (int i = 0; i < size; i++)
	ifs.read((char *) &v[i], sizeof(double));
      fv.push_back(v);
    }

    ifs.close();
  }

  return fv;
}

