/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    visualizeDetections.cpp
** AUTHOR(S):   Olga Russakovsky <olga@cs.stanford.edu>
**              Ian Goodfellow <ia3n@cs.stanford.edu>
** DESCRIPTION:
**   Given an XML file and a sequence of images, displays the images with the 
**   detected objects outlined.
*****************************************************************************/

#include <cstdlib>
#include <string>
#include <limits>
#include <iomanip>
#include <fstream>
#include <map>
#include <vector>

#include "svlBase.h"
#include "svlVision.h"

using namespace std;

#define WINDOW_NAME "detections"
#define MAX_CHAR 255

void usage()
{
    cerr << SVL_USAGE_HEADER << endl;
    cerr << "USAGE: ./visualizeDetections [OPTIONS] <objects xml> <imageSequence>" << endl;
    cerr << "OPTIONS:" << endl
	 << "  -t <num>    :: threshold for displaying objects (default: 0.5)" << endl
	 << " -o <dirname> :: instead of displaying images, dump them to a directory and make an html file to display them"
	 << " -nm          :: use non-maxima suppression"
	 << endl;
}

static void renderImage(IplImage * img, const svlObject2dFrame & objects, double threshold)
{
  int num = 0;
  for (size_t j = 0; j < objects.size(); j++)
    if (objects[j].pr > threshold)
      num++;

  int numToShow = min(num, 200);
  cerr << "\tShowing " << numToShow << " object(s)";
  if (numToShow < num)
    cerr << " (out of " << num << " originally)";
  cerr << endl;
  
  int numShown = 0;
  for (unsigned j = 0; j < objects.size(); j++) {
    const svlObject2d &obj = objects[j];
    if (obj.pr > threshold) {
      int x1 = (int)(obj.x);
      int y1 = (int)(obj.y);
      int x2 = (int)(obj.x + obj.w - 1);
      int y2 = (int)(obj.y + obj.h - 1);
      
      cvRectangle(img, cvPoint(x1, y1), cvPoint(x2, y2),
		  CV_RGB(0, 255, 0), 1);
      numShown++;
      if (numShown > numToShow) break;
    }
  }
}


static int showImage(IplImage *&img) {

  
  cvShowImage(WINDOW_NAME, img);
  return cvWaitKey(0) & MAX_CHAR;
}



#define NUM_REQUIRED_PARAMETERS 2

int main(int argc, char **argv)
{
  char *inputFilename;
  char *imageSeq;
  char * outdir = NULL;
  string outdirStr;
  double threshold = 0.5;
  bool bUseNonmax = false;

  cvNamedWindow(WINDOW_NAME, 0);
  
  char **args = argv + 1;
  while (--argc > NUM_REQUIRED_PARAMETERS) {
    if (!strcmp(*args, "-t")) {
      threshold = atof(*(++args)); argc--;
    } 
    else if (!strcmp(*args, "-o"))
      {
	args++;
	argc--;

	outdir = *args;
      }
    else if (!strcmp(*args, "-nm"))
      {
	bUseNonmax = true;
      }
    else if (**args == '-') {
      SVL_LOG(SVL_LOG_FATAL, "unrecognized option " << *args);
    } else {
      break;
    }
    args++;
  }

  if (argc < NUM_REQUIRED_PARAMETERS) {
    usage();
    return -1;
  }

  inputFilename = args[0];
  imageSeq = args[1];

  ofstream ofs;

  if (outdir)
    {
      outdirStr = string(outdir);

      svlCreateDirectory(outdir);

      ofs.open((outdirStr+"/index.html").c_str());

      ofs << "<html> \n"
	  << "<body> \n";

      assert(!ofs.fail());
    }


  svlObject2dSequence objects;
  SVL_LOG(SVL_LOG_VERBOSE, "Reading from " << inputFilename);
	
  if (!readObject2dSequence(inputFilename, objects)) {
    SVL_LOG(SVL_LOG_ERROR, "...failed to read " << inputFilename);
  }

  if (bUseNonmax)
    {
      svlObjectDetectionAnalyzer d(true);
      d.loadDetections(objects); 
      objects = d.getDetections();
    }

  svlImageSequence sequence;
  SVL_ASSERT_MSG(strExtension(imageSeq).compare("xml") == 0, imageSeq << " is not an image sequence");
  sequence.load(imageSeq);
 
  IplImage *img = NULL;
  int level = 0;
  int numLevels = sequence.size();
  svlObject2dFrame tmp;
  while (true) {
    cerr << "Showing level " << level+1 << " of " << numLevels << endl;
    // get the correct image at the correct level (and make it non-const)
    img = cvCloneImage(sequence.image(level));
    // find the corresponding list of objects

    map<string, svlObject2dFrame>::iterator iter = objects.find(strBaseName(sequence[level]));

    int key = -1;
    if (iter == objects.end())
      {
	if (!outdir)
	  key = showImage(img);
      }
    else 
      {
	renderImage(img, iter->second, threshold);
	if (!outdir)
	  key = showImage(img );
      }


    
    if (outdir)
      {

	string filename = strBaseName(sequence[level])+".jpg";

	stringstream s;
	s << outdirStr << "/" << filename;

	cvSaveImage(s.str().c_str(), img);

	ofs << "<img src=\" "
	    << filename
	    << "\" > \n";


	if (level < numLevels - 1)
	  level++;
	else
	  break;
      }
    else if (key == (int)'n') {
      if (level < numLevels-1) level++;
    } else if (key == (int)'b') {
      if (level > 0) level--;
    } else if (key == (int)'s') {
      string name = sequence.directory() + string("/visualize_") + sequence[level];
      cvSaveImage(name.c_str(), img);
    } else if (key == (int)'q') {
      break;
    }
  }


  if (outdir)
    {
      ofs << "</body> \n"
	  << "</html> \n";
      ofs.close();
    }

  return 0;
}

