/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlPlaneClustering.cpp
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
**
** DESCRIPTION: Switchyard class for performing plane clustering.
**
*****************************************************************************/

#if defined(_WIN32)||defined(WIN32)||defined(__WIN32__)
#define _CRT_SECURE_NO_DEPRECATE
#undef min
#undef max
#endif

#include "svlPlaneClustering.h"

#define VALID_MAT_INDEX(pt,mat) pt.x>=0&&pt.y>=0&&pt.x<mat->cols&&pt.y<mat->rows
#define VALID_MAT_INDEX_PAIR(pt,mat) pt.first>=0&&pt.second>=0&&pt.first<mat->cols&&pt.second<mat->rows

#if 0
int svlPlaneClustering::mainloop() {}
void svlPlaneClustering::flow_atom_received(SYFlow *flow) {}
#endif

svlPlaneClustering::svlPlaneClustering() {
	init();
	w = 640;
	h = 480;
}

svlPlaneClustering::svlPlaneClustering(int w, int h) : w(w), h(h) {
	init();
}

svlPlaneClustering::svlPlaneClustering(int w, int h, svlCameraIntrinsics intrinsics, svlCameraExtrinsics extrinsics) : w(w), h(h) {
	init();
	setInExtrinsics(w,h,intrinsics,extrinsics);
	/**_intrinsics = svlCameraIntrinsics(intrinsics);
	*_extrinsics = svlCameraExtrinsics(extrinsics);
	double scale = (double)(_ZHat->cols)/640.0;
	if ( fabs(scale-1.0) > 1e-3 ) {
		_intrinsics->rescale(scale);
	}*/
}

void svlPlaneClustering::init() {
	_reconstP = new vector<svlPoint3d>();
	_reconstN = new vector<svlPoint3d>();
	_intrinsics = new svlCameraIntrinsics();
	_extrinsics = new svlCameraExtrinsics();
	_regSizes.setAscending(false);
	declareDefaultOptions();
}

svlPlaneClustering::~svlPlaneClustering() {
	_reconstP->clear();
	_reconstN->clear();
	delete _reconstP;
	delete _reconstN;
	delete _intrinsics;
	delete _extrinsics;
}

void svlPlaneClustering::setInExtrinsics(int w_, int h_, svlCameraIntrinsics intrinsics, svlCameraExtrinsics extrinsics) {
	w = w_;
	h = h_;
	*_intrinsics = svlCameraIntrinsics(intrinsics);
	*_extrinsics = svlCameraExtrinsics(extrinsics);
}

void svlPlaneClustering::setCvMats(CvMat* X, CvMat* Y, CvMat* Z, CvMat* nX, CvMat* nY, CvMat* nZ, CvMat* Ix) {
	_X = X;
	_Y = Y;
	_Z = Z;
	_nX = nX;
	_nY = nY;
	_nZ = nZ;
	_Ix = Ix;
}

void svlPlaneClustering::declareDefaultOptions() {
    declareOption("ClusteringType", ( 1 ? "k-means" : "Rabbani" ));
    declareOption("k-means k", "10");
    declareOption("% of max # pts", "0.1");
}

void svlPlaneClustering::computeHulls(vector<svlPoint3d> &Pts) {
	// 3D and 2D projection matrices
	CvMat *P = cvCreateMat(3,_regSizes[0].first,CV_32FC1);
	CvMat *P2 = cvCreateMat(2,P->cols,CV_32FC1);
	// SVD matrices
	CvMat *PPT = cvCreateMat(3,3,CV_32FC1);
	CvMat *U = cvCreateMat(3,3,CV_32FC1); // left orthogonal matrix
	CvMat *S = cvCreateMat(3,3,CV_32FC1); // singular values
	CvMat *l = cvCreateMat(3,1,CV_32FC1); // vector
	CvMat *U12 = cvCreateMat(2,3,CV_32FC1); // projection matrix
	CvMat *minmax = cvCreateMat(2,2,CV_32FC1);
	CvMemStorage *storage = cvCreateMemStorage(100);

	int j;
#define HULLS_DEBUG_OUT 0
#if HULLS_DEBUG_OUT
	FILE *fo; int foi;
	/*fo = fopen("zzz_hulls.out","wb");
	fwrite(&(foi=(int)Pts.size()),sizeof(int),1,fo);
	fwrite(&(foi=(int)_planes.size()),sizeof(int),1,fo);
	fwrite(&(foi=(int)_regionIx.size()),sizeof(int),1,fo);
	fwrite(&(foi=(int)_regSizes.size()),sizeof(int),1,fo);
	for ( int i=0; i<(int)_regSizes.size(); ++i ) {
		fwrite(&(foi=_regSizes[i].first),sizeof(int),1,fo);
		fwrite(&(foi=_regSizes[i].second),sizeof(int),1,fo);
	}
	fclose(fo);*/
	fo = fopen("zzz_hulls.out","wb");
#endif
	for ( int i=0; i<(int)_regSizes.size(); ++i ) {
		/*fwrite(&(foi=-1),sizeof(int),1,fo);
		fwrite(&(foi=_regSizes[i].second),sizeof(int),1,fo);
		fwrite(&(foi=(int)_regionIx[_regSizes[i].second].size()),sizeof(int),1,fo);*/
		//fwrite(&j,sizeof(int),1,fo);
		if ( (j=_reg2plane.getAtoB(_regSizes[i].second)) == -1 )
			break; // get the region index, convert to plane
		_planes[j].computeHulls(Pts, _regionIx[_regSizes[i].second], P, P2, PPT, U, S, l, U12, minmax, storage);
#if HULLS_DEBUG_OUT
		fwrite(&_planes[j].p.x,sizeof(double),1,fo);
		fwrite(&_planes[j].p.y,sizeof(double),1,fo);
		fwrite(&_planes[j].p.z,sizeof(double),1,fo);
		fwrite(&_planes[j].n.x,sizeof(double),1,fo);
		fwrite(&_planes[j].n.y,sizeof(double),1,fo);
		fwrite(&_planes[j].n.z,sizeof(double),1,fo);
		fwrite(&(foi=(int)_planes[j].Hull3d.size()),sizeof(int),1,fo);
		for ( int k=0; k<foi; ++k ) {
			fwrite(&_planes[j].Hull3d[k].x,sizeof(double),1,fo);
			fwrite(&_planes[j].Hull3d[k].y,sizeof(double),1,fo);
			fwrite(&_planes[j].Hull3d[k].z,sizeof(double),1,fo);
		}
#endif
	}
#if HULLS_DEBUG_OUT
	fclose(fo);
#endif

	cvReleaseMat(&P);
	cvReleaseMat(&P2);
	cvReleaseMat(&PPT);
	cvReleaseMat(&U);
	cvReleaseMat(&S);
	cvReleaseMat(&l);
	cvReleaseMat(&U12);
	cvReleaseMat(&minmax);
	cvReleaseMemStorage(&storage);
}

vector<svlPlane> svlPlaneClustering::computeHulls(vector<svlPoint3d> &Pts, vector<vector<int> > &regionIx) {
	// 3D and 2D projection matrices
	// find max-needed point cloud size
	int max_size = (int)regionIx[0].size();
	for ( int i=1; i<(int)regionIx.size(); ++i ) {
		max_size = (int)regionIx[i].size() > max_size ? (int)regionIx[i].size() : max_size;
	}
	CvMat *P = cvCreateMat(3,max_size,CV_32FC1);
	CvMat *P2 = cvCreateMat(2,P->cols,CV_32FC1);
	// SVD matrices
	CvMat *PPT = cvCreateMat(3,3,CV_32FC1);
	CvMat *U = cvCreateMat(3,3,CV_32FC1); // left orthogonal matrix
	CvMat *S = cvCreateMat(3,3,CV_32FC1); // singular values
	CvMat *l = cvCreateMat(3,1,CV_32FC1); // vector
	CvMat *U12 = cvCreateMat(2,3,CV_32FC1); // projection matrix
	CvMat *minmax = cvCreateMat(2,2,CV_32FC1);
	CvMemStorage *storage = cvCreateMemStorage(100);

	vector<svlPlane> ret(regionIx.size(),svlPlane());
	for ( int i=0; i<(int)regionIx.size(); ++i ) {
		ret[i].computeHulls(Pts, regionIx[i], P, P2, PPT, U, S, l, U12, minmax, storage);
	}

	cvReleaseMat(&P);
	cvReleaseMat(&P2);
	cvReleaseMat(&PPT);
	cvReleaseMat(&U);
	cvReleaseMat(&S);
	cvReleaseMat(&l);
	cvReleaseMat(&U12);
	cvReleaseMat(&minmax);
	cvReleaseMemStorage(&storage);

	return ret;
}

void svlPlaneClustering::computeCameraPlanes(bool do_hulls) {
	//assert(_extrinsics!=NULL);
#define OUT_CAM_HULLS 0
#if OUT_CAM_HULLS
	FILE *fo; int foi;
	fo = fopen("zzz_cam_hulls.out","wb");
#endif
	_camPlanes.clear();
	_camPlanes.reserve(_planes.size());
	//_camPlanes = vector<svlPlane>((int)_planes.size(), svlPlane());
	//_camPlanes.resize((int)_planes.size());
	for ( int i=0; i<(int)_planes.size(); ++i ) {
		_camPlanes.push_back(svlPlane());
		_camPlanes.back().p = _extrinsics->world2camera_pt(_planes[i].p);
		_camPlanes.back().n = _extrinsics->worldRot2camera_pt(_planes[i].n);
		_camPlanes.back().off = _camPlanes.back().n.dot(_camPlanes.back().p);
		_camPlanes.back().u1 = _extrinsics->worldRot2camera_pt(_planes[i].u1);
		_camPlanes.back().u2 = _extrinsics->worldRot2camera_pt(_planes[i].u2);
		_camPlanes.back().min_pt2 = _planes[i].min_pt2;
		_camPlanes.back().max_pt2 = _planes[i].max_pt2;
		if ( do_hulls ) {
			double gd, gdmax = 0.0;
			_camPlanes.back().Hull3d.reserve(_planes[i].Hull3d.size());
			for ( int j=0; j<(int)_planes[i].Hull3d.size(); ++j ) {
				_camPlanes.back().Hull3d.push_back( _extrinsics->world2camera_pt( _planes[i].Hull3d[j] ) );
			}
#if 0
			_camPlanes.back().HullPs = _planes[i].HullPs;
			_camPlanes.back().hullLims = _planes[i].hullLims;
#else
			_camPlanes.back().HullPs.reserve(_planes[i].HullPs.size());
			//_camPlanes.back().hullLims = _planes[i].hullLims;
			_camPlanes.back().hullLims.reserve(_planes[i].hullLims.size());
			for ( int j=0; j<(int)_planes[i].HullPs.size(); ++j ) {
				_camPlanes.back().HullPs.push_back( _extrinsics->worldRot2camera_pt( _planes[i].HullPs[j] ) );
				gdmax = _camPlanes.back().HullPs.back().dot(_camPlanes.back().Hull3d.front());
				for ( int k=1; k<(int)_camPlanes.back().Hull3d.size(); ++k )
					if ( (gd=_camPlanes.back().HullPs.back().dot(_camPlanes.back().Hull3d[k] - _camPlanes.back().Hull3d[j])) > gdmax )
						gdmax = gd;
				_camPlanes.back().hullLims.push_back(gdmax);
			}
#endif
		}
#if OUT_CAM_HULLS
		fwrite(&_camPlanes.back().p.x,sizeof(double),1,fo);
		fwrite(&_camPlanes.back().p.y,sizeof(double),1,fo);
		fwrite(&_camPlanes.back().p.z,sizeof(double),1,fo);
		fwrite(&_camPlanes.back().n.x,sizeof(double),1,fo);
		fwrite(&_camPlanes.back().n.y,sizeof(double),1,fo);
		fwrite(&_camPlanes.back().n.z,sizeof(double),1,fo);
		fwrite(&_camPlanes.back().off,sizeof(double),1,fo);
		fwrite(&(foi=(int)_camPlanes.back().Hull3d.size()),sizeof(int),1,fo);
		for ( int k=0; k<foi; ++k ) {
			fwrite(&_camPlanes.back().Hull3d[k].x,sizeof(double),1,fo);
			fwrite(&_camPlanes.back().Hull3d[k].y,sizeof(double),1,fo);
			fwrite(&_camPlanes.back().Hull3d[k].z,sizeof(double),1,fo);
		}
#endif
	}
#if OUT_CAM_HULLS
	fclose(fo);
#endif
}

void svlPlaneClustering::outputCameraPlanes(FILE* fout, bool do_hulls) {
	assert(fout!=NULL);
	int foi = (int)_camPlanes.size();
	fwrite(&foi,sizeof(int),1,fout);
	for ( int i=0; i<(int)_camPlanes.size(); ++i ) {
		fwrite(&_camPlanes[i].p.x,sizeof(double),1,fout);
		fwrite(&_camPlanes[i].p.y,sizeof(double),1,fout);
		fwrite(&_camPlanes[i].p.z,sizeof(double),1,fout);
		fwrite(&_camPlanes[i].n.x,sizeof(double),1,fout);
		fwrite(&_camPlanes[i].n.y,sizeof(double),1,fout);
		fwrite(&_camPlanes[i].n.z,sizeof(double),1,fout);
		fwrite(&_camPlanes[i].off,sizeof(double),1,fout);
		if ( do_hulls ) {
			fwrite(&(foi=(int)_camPlanes[i].Hull3d.size()),sizeof(int),1,fout);
			for ( int k=0; k<(int)_camPlanes[i].Hull3d.size(); ++k ) {
				fwrite(&_camPlanes[i].Hull3d[k].x,sizeof(double),1,fout);
				fwrite(&_camPlanes[i].Hull3d[k].y,sizeof(double),1,fout);
				fwrite(&_camPlanes[i].Hull3d[k].z,sizeof(double),1,fout);
			}
		}
	}
}

bool svlPlaneClustering::segmentSuperPointCloud(svlPointCloudData *pcData, CvMat *Ix) {
#define SEGSUPER_DEBUG_OUT 0
#if SEGSUPER_DEBUG_OUT
	FILE *fo; int foi; double fod; char c[32];
	static int n_super_rounds;
#endif
	//if ( (int)pcData->pointCloud.size() < _Ix->rows*_Ix->cols ) return false;
	int n_pts = (int)pcData->pointCloud.size();

	string clusteringType = getOption("ClusteringType");
	if ( true || clusteringType.compare("k-means") == 0 ) { // do k-means clustering
		int k = getOptionAsInt("k-means k");
		//_planes = vector<svlPlane>();//(k,svlPlane());
		//cerr << "Running k-means with " << k << " centroids..." << endl;
		assert(k>0);
		svlKMeansClustering kmc;
		kmc.setOption("% of max # pts","0.1");
#if SEGSUPER_DEBUG_OUT
		sprintf(c,"yyy_kplanes_%d_%d_in.out",n_super_rounds,(int)_planes.size()); fo = fopen(c,"wb"); fclose(fo);
#endif
		kmc.cluster(*pcData, _ix, _regionIx, _planes, k);

		// keep only largest planes and flattest regions
		kmc.populateRegSizes(_regionIx,_regSizes);
		_regSizes.sort();
		_ix = vector<int>(n_pts,-1); // init point assignments
		int new_ix = 0;
		_reg2plane.clear();
		for ( int i=0; i<(int)_regSizes.size() && _regSizes[i].first>=1000; ++i ) {
			int old_ix = (int)_regSizes.ix_at(i);
			//_reg2plane.addAtofromB(old_ix,new_ix);
			_reg2plane.addAtofromB(i,i);
			if ( !_planes[old_ix].isHorizontal() && !_planes[old_ix].isVertical() )
				continue;
			_planes.push_back(_planes[old_ix]); // copy this plane onto the end of the vector
			// ... and do things to the copy
			for ( vector<int>::const_iterator it=_regionIx[old_ix].begin(); it!=_regionIx[old_ix].end(); ++it ) {
				// move points along to new cluster only if they are within 5 cm of the plane
				if ( fabs(_planes.back().perp_distance(pcData->pointCloud[*it])) <= 0.05 ) {
					_ix[*it] = new_ix; // promote to new plane indices out of plane
				} // else leave as -1
			}
			++new_ix;
		}
		_planes.erase(_planes.begin(), _planes.begin()+k); // get rid of the first k planes
		kmc.populateRegionIxAndSizes(_ix,new_ix,_regionIx,_regSizes);
		_regSizes.sort();
#if SEGSUPER_DEBUG_OUT
		fclose(fo);
		sprintf(c,"yyy_kplanes_%d_%d_out.out",n_super_rounds++,(int)_planes.size()); fo = fopen(c,"wb"); fclose(fo);
#endif

		computeHulls(pcData->pointCloud);

	} else if ( clusteringType.compare("Rabbani") == 0 ) { // do Rabbani clustering
		// calculate residuals from images
	        //svlSortVector<double,int> resid(true); // ascending sort
	        svlSortVectorDI resid(true); // ascending sort
		//resid.reserve((int)pcData->pointCloud.size());
		resid._myvec = vector<pair<double,int> >(n_pts, make_pair(1e9,-1));
		int gi, gi1, gi2, inc_step = 2; // +/- n_inc * inc_step
		svlPoint3d p, n;
		for ( int r=0; r<Ix->rows; ++r ) {
			for ( int c=0; c<Ix->cols; ++c ) {
				if ( (gi=CV_MAT_ELEM(*Ix,int,r,c)) >= 0 && gi < resid.size() ) {
					// calculate resid for this point
					p = 0.0;
					n = 0.0; // zeralize point accumulators
					gi2 = 0;
					for ( int ro=max(0,r-(inc_step<<1)); ro<=min(Ix->rows-1,r+(inc_step<<1)); ro+=inc_step ) {
						for ( int co=max(0,c-(inc_step<<1)); co<=min(Ix->cols-1,c+(inc_step<<1)); co+=inc_step ) {
							if ( (gi1=CV_MAT_ELEM(*Ix,int,ro,co)) >= 0 && gi1 < resid.size() ) {
								p += pcData->pointCloud[gi1];
								n += pcData->pointNormals[gi1];
								++gi2;
							}
						}
					}
					if ( gi2 > 0 && fabs(n.norm()) >= 1e-3 ) {
						p /= (double)gi2;
						n /= n.norm();
						resid._myvec[gi].first = n.dot(pcData->pointCloud[gi]) - n.dot(p); // residual
						resid._myvec[gi].second = gi; // index
					}
				}
			}
		}
#if SEGSUPER_DEBUG_OUT && 0
		fo = fopen("yyy_resid.out","wb");
		for ( int i=0; i<(int)resid.size(); ++i ) {
			fwrite(&resid._myvec[i].first,sizeof(double),1,fo);
			fwrite(&resid._myvec[i].second,sizeof(int),1,fo);
		}
		fclose(fo);
#endif
#if SEGSUPER_DEBUG_OUT && 0
		fo = fopen("yyy_ix_out.out","wb");
		for ( int r=0; r<(int)Ix->rows; ++r ) {
			for ( int c=0; c<(int)Ix->cols; ++c ) {
				foi = CV_MAT_ELEM(*Ix,int,r,c);
				fwrite(&foi,sizeof(int),1,fo);
			}
		}
		fclose(fo);
#endif

		cerr << "Running Rabbani..." << endl;
		svlRabbaniClustering rpcc;
		rpcc.cluster(*pcData,Ix,_ix,_regionIx,resid);
	} else {
		cerr << "Running nothing..." << endl;
	}

#if SEGSUPER_DEBUG_OUT
	// output _reg2plane contents for debugging
	fo = fopen("yyy_reg2plane.out","wb");
	fwrite(&(foi=(int)_reg2plane.AtoB.size()),sizeof(int),1,fo);
	fwrite(&(foi=-1),sizeof(int),1,fo);
	_reg2plane.setIterABegin();
	while ( !_reg2plane.iterAEnd() ) {
		fwrite(&(foi=_reg2plane.getIterA().first),sizeof(int),1,fo);
		fwrite(&(foi=_reg2plane.getIterA().second),sizeof(int),1,fo);
		_reg2plane.incIterA();
	}
	fwrite(&(foi=-1),sizeof(int),1,fo);
	_reg2plane.setIterBBegin();
	while ( !_reg2plane.iterBEnd() ) {
		fwrite(&(foi=_reg2plane.getIterB().first),sizeof(int),1,fo);
		fwrite(&(foi=_reg2plane.getIterB().second),sizeof(int),1,fo);
		_reg2plane.incIterB();
	}
	fwrite(&(foi=-1),sizeof(int),1,fo);
	for ( int i=0; i<(int)_regSizes.size(); ++i ) {
		fwrite(&(foi=_regSizes[i].first),sizeof(int),1,fo);
		fwrite(&(foi=_regSizes[i].second),sizeof(int),1,fo);
	}
	fclose(fo);
#endif

	return false;
}

bool svlPlaneClustering::segmentRawPointCloud(vector<svlPoint3d> *inP) {
	// kd-tree the points; generate normals
#define SEGRAW_DEBUG_OUT 0

	if ( (int)inP->size() < 1e4 )
		return false;

#if SEGRAW_DEBUG_OUT
	FILE *fo; int foi; double fod;
#endif

#if 1
	// create decimated (reconstructed) point cloud
	_reconstP->clear();
	svlKdTree *xdt = new svlKdTree(inP, // use the raw point cloud
				  0.05, // allow no box of size DxDxD m^3 to contain more than 1 point
				  _reconstP, // output into reconstructed point cloud
				  20); // once done, window-shade back up to bucket size of 20
#else
	svlKdTree *xdt = new svlKdTree(inP,20); // or no reconstruction
	_reconstP = inP;
#endif

	// get all nearest neighbors in one go
	vector<vector<int> > allNN;
	xdt->findAllKNN(20,0.1,allNN);
	delete xdt;

	int n_pts = (int)_reconstP->size();
	_reconstN->resize(n_pts); // allocate normals

	//svlSortVector<double, int> resid(n_pts);
	svlSortVectorDI resid(n_pts);
	_sva.computeNNNandRes(*_reconstP, *_reconstN, allNN, &resid);

	//fo = fopen("yyy_NNN.out","wb"); fclose(fo);

#if SEGRAW_DEBUG_OUT // write out points and normals for debugging
	fo = fopen("zyz_PN_dec.out","wb");
	fwrite(&n_pts,sizeof(int),1,fo);
	for ( int i=0; i<n_pts; ++i ) {
		fwrite(&(*_reconstP)[i].x,sizeof(double),1,fo);
		fwrite(&(*_reconstP)[i].y,sizeof(double),1,fo);
		fwrite(&(*_reconstP)[i].z,sizeof(double),1,fo);
	}
	for ( int i=0; i<n_pts; ++i ) {
		fwrite(&(*_reconstN)[i].x,sizeof(double),1,fo);
		fwrite(&(*_reconstN)[i].y,sizeof(double),1,fo);
		fwrite(&(*_reconstN)[i].z,sizeof(double),1,fo);
	}
	for ( int i=0; i<n_pts; ++i ) {
		fwrite(&(resid[i].first),sizeof(double),1,fo);
	}
	fclose(fo);
#endif

	// cluster with Rabbani
	_ix = vector<int>(n_pts,-1); // allocate assignments vector before sending to the clusterers
	_regionIx.clear(); // clear old regions ...
	_planes.clear(); // ... and planes
	svlRabbaniClustering rpcc;
	rpcc.cluster(*_reconstP, *_reconstN, allNN, _ix, _regionIx, resid, 3); // mode 2
	// take largest regions and push to planes: desceding sort on sizes
	rpcc.populateRegSizes(_regionIx,_regSizes);
	_regSizes.sort();
	// construct svlPlanes out of largest regions

#if SEGRAW_DEBUG_OUT
	fo = fopen("yyy_rab_ix.out","wb");
	for ( int i=0; i<(int)_ix.size(); ++i ) {
		fwrite(&_ix[i],sizeof(int),1,fo);
	}
	fclose(fo);
#endif

#define MIN_PLANE_PTS 50
	_planes.clear();
	_reg2plane.clear();
	for ( int i=0; i<(int)_regSizes.size(); ++i ) {
		if ( _regSizes[i].first < MIN_PLANE_PTS )
			break;
		_reg2plane.addAtofromB(_regSizes[i].second,(int)_planes.size());
		_planes.push_back(svlPlane());
	}

	computeHulls(*_reconstP);
	
	computeCameraPlanes(true); // with Hulls

	return true; // for now, return no clustering results
}

inline svlPoint2d pt2nearestImgPt(svlPoint2d &pt, CvMat *img) {
	svlPoint2d ret;
	double gd = pt.x > img->cols-1 ? img->cols-1 : pt.x;
	ret.x = gd < 0 ? 0 : gd;
	gd = pt.y > img->rows-1 ? img->rows-1 : pt.y;
	ret.y = gd < 0 ? 0 : gd;
	return ret;
	/*return svlPoint2d(min(0, max( pt.x, img->cols-1 )),
					  min(0, max( pt.y, img->rows-1 )));*/
}

void svlPlaneClustering::drawHulls(CvMat *Hulls) {
#define DRAWHULLS_DEBUG_OUT 0
	cvZero(Hulls);

	int gi;
	svlPoint3d pt1, pt2, ray;
	pair<int,int> pair1, pair2, acc;
	svlPoint2d pp1, pp2, deriv;

	// find proper order to process _camPlanes in for "painter's algorithm"
	svlSortVector<double,int> cam_dist((int)_camPlanes.size(),false); // descending
	for ( int i=0; i<(int)_camPlanes.size(); ++i ) {
		cam_dist.push_back(_camPlanes[i].p.norm2());
	}
	cam_dist.sort();
#if DRAWHULLS_DEBUG_OUT
	fo = fopen("yyy_engirt.out","wb");
#endif
	bool drew;
	for ( int ix=0; ix<(int)_camPlanes.size(); ++ix ) {
		int i = cam_dist.ix_at(ix);
		if ( (int)_camPlanes[i].Hull3d.size() <= 1 ) continue;
#if DRAWHULLS_DEBUG_OUT
		fwrite(&i,sizeof(int),1,fo);
#endif
		pair1 = _intrinsics->point(_camPlanes[i].Hull3d.back(),true); // projected index
		pp1 = svlPoint2d(pair1.first,pair1.second); // x, y
		drew = false;
#if DRAWHULLS_DEBUG_OUT
		if ( i == 0 ) {
			fwrite(&pp1.x,sizeof(double),1,fo);
			fwrite(&pp1.y,sizeof(double),1,fo);
		}
#endif
		for ( int j=0; j<(int)_camPlanes[i].Hull3d.size(); ++j ) {
			pair2 = _intrinsics->point(_camPlanes[i].Hull3d[j],true); // projected index			
			pp2 = svlPoint2d(pair2.first,pair2.second); // x, y
#if DRAWHULLS_DEBUG_OUT
			if ( i == 0 ) {
				fwrite(&pp2.x,sizeof(double),1,fo);
				fwrite(&pp2.y,sizeof(double),1,fo);
			}
#endif
			if ( !( ( pp1.x < 0 && pp2.x < 0 )
				|| ( pp1.x >= Hulls->cols && pp2.x >= Hulls->cols )
				|| ( pp1.y < 0 && pp2.y  < 0 )
				|| ( pp1.y >= Hulls->rows && pp2.y >= Hulls->rows ) ) ) { // easy out-of-bounds checks
				deriv = pp2-pp1; // draw from 1 to 2
				
				if ( VALID_MAT_INDEX(pp1,Hulls) ) {
					drew = true;
					CV_MAT_ELEM(*Hulls,float,(int)pp1.y,(int)pp1.x) = (float)i;
				} // _plane i
				if ( fabs(deriv.x) > fabs(deriv.y) ) { // increment in x
					gi = (int)fabs(deriv.x); // # of steps to go
					deriv /= fabs(deriv.x); // normalize the derivative to count as one increment
				} else { // increment in y
					gi = (int)fabs(deriv.y);
					deriv /= fabs(deriv.y);
				}
				for ( int m=0; m<gi; ++m ) {
					pp1 += deriv;
					if ( VALID_MAT_INDEX(pp1,Hulls) ) {
						drew = true;
						CV_MAT_ELEM(*Hulls,float,(int)pp1.y,(int)pp1.x) = (float)i;
					}
				}
				if ( VALID_MAT_INDEX(pp2,Hulls) ) {
					drew = true;
					CV_MAT_ELEM(*Hulls,float,(int)pp2.y,(int)pp2.x) = (float)i;
				}// _plane i
			}
			pp1 = pp2;
		}
		if ( drew && false ) { // floodfill using center point
			pair2 = _intrinsics->point(_camPlanes[i].p,true); // center point for filling
			svlPoint2d tmpPt(pair2.first,pair2.second);
			pp1 = pt2nearestImgPt(tmpPt, Hulls);
			CvPoint cpt;
			cpt.x = (int)pp1.x;
			cpt.y = (int)pp1.y;
			cvFloodFill(Hulls,cpt,cvRealScalar((double)i), cvScalarAll(0), cvScalarAll(0));
		}
	}
#if DRAWHULLS_DEBUG_OUT
	fclose(fo);
#endif
}

void svlPlaneClustering::computeZHat(CvMat *ZHat) {
	// back-project indices onto image
#define CZH_DEBUG_OUT 0
#if CZH_DEBUG_OUT
	FILE *fo; int foi; double fod; float fof;
	// draw and output hulls
	CvMat *H = cvCreateMat(ZHat->rows,ZHat->cols,CV_32FC1);
	cvZero(H);
	drawHulls(H);
	fo = fopen("yyy_Hulls_out.out","wb");
	for ( int c=0; c<H->cols; ++c ) {
		for ( int r=0; r<H->rows; ++r ) {
			foi = CV_MAT_ELEM(*H,int,r,c);
			fwrite(&foi,sizeof(int),1,fo);
		}
	}
	fclose(fo);
	cvReleaseMat(&H);
#endif

	if ( w != ZHat->cols ) {
		double scale = (double)(ZHat->cols)/(double)w;
		if ( fabs(scale-1.0) > 1e-3 ) {
			_intrinsics->rescale(scale);
		}
		w = ZHat->cols;
		h = ZHat->rows;
	}

	// use a k-d tree instead of the projection matrix and ... things
	// push index points into templated k-d tree
	CvMat* Ix1 = cvCreateMat(ZHat->rows,ZHat->cols,CV_32SC1); // indices matrix, same size as ZHat
	CvMat* Gt = cvCreateMat(ZHat->rows,ZHat->cols,CV_32SC1); // indices matrix, same size as ZHat
	cvZero(Gt);
	cvZero(Ix1);

	pair<int,int> pt2;
	svlPoint3d proj_pt;
	vector<svlPoint2d> P2;
	vector<int> P22plane, P22ix;
	// project points onto image (?) to generate Ix1 and fill back-permutation vectors P22plane, P22ix ...
	for ( int i=0; i<(int)_ix.size(); ++i ) {
		pt2 = _intrinsics->point(_extrinsics->world2camera((*_reconstP)[i]),true); // projected index
		if ( VALID_MAT_INDEX_PAIR(pt2,Ix1) ) {
			P2.push_back(svlPoint2d((double)pt2.second,(double)pt2.first));
			P22ix.push_back(i);
			P22plane.push_back( _reg2plane.getAtoB(_ix[i]) );
			// write as ground truth into Ix1 (not anymore)
			CV_MAT_ELEM(*Gt,int,pt2.second,pt2.first) = P22plane.back()+1;//_reg2plane.getAtoB(_ix[i])+1//reg2plane[_ix[i]+1];
		}
	}
	svlKdTreeT<svlPoint2d> kdt(&P2,20); // bucket size of 20

	// fill ZHat
	int gi, k = 8;
	double search_radius = (double)(ZHat->cols)/20.0; // -1 for infinite
	//svlPointNd<int,2> srch_pt;
	svlPoint2d srch_pt;
	vector<int> NN;
	map<int,pair<int,double> > planesDCtr;
	for ( int r=0; r<Ix1->rows; ++r ) {
		for ( int c=0; c<Ix1->cols; ++c ) {
			svlPoint3d ray = _intrinsics->ray_pt((double)c,(double)r);
			if ( (gi=-1+CV_MAT_ELEM(*Ix1,int,r,c)) >= 0 ) {
				ray *= _camPlanes[gi].interceptScale(ray);
				CV_MAT_ELEM(*ZHat,float,r,c) = (float)ray.z;//ray.norm();
			} else { // check NN
				srch_pt.x = (double)r;
				srch_pt.y = (double)c;
				NN.clear();
				kdt.findKNN(srch_pt,k,search_radius,NN);
				if ( (int)NN.size() > 0 ) {
					int max_ct = -1, max_plane = -1;
					planesDCtr.clear();
					for ( vector<int>::iterator i=NN.begin(); i!=NN.end(); ++i ) {
						++(planesDCtr[ P22plane[*i] ].first);
						planesDCtr[ P22plane[*i] ].second += (*_reconstP)[ P22ix[*i] ].norm();
					}
					// of the represented planes, select the nearest
					for ( map<int,pair<int,double> >::iterator it=planesDCtr.begin(); it!=planesDCtr.end(); ++it )
						it->second.second /= (double)it->second.first; // normalize norm sums
					map<int,pair<int,double> >::iterator max_it = planesDCtr.begin();
					for ( map<int,pair<int,double> >::iterator it=planesDCtr.begin(); it!=planesDCtr.end(); ++it )
						if ( it->second.second < max_it->second.second )
							max_it = it;
					max_plane = max_it->first;
					max_ct = max_it->second.first;

					// CHEAT: Fill in advance window
					if ( max_plane >= 0 ) { //&& max_ct >= 2 ) 
						ray *= _camPlanes[max_plane].interceptScale(ray);
						// check that it resides within relevant hull first
						if ( !_camPlanes[max_plane].insideHull(ray) )
							continue;
						CV_MAT_ELEM(*ZHat,float,r,c) = (float)ray.z;//ray.norm();
						++max_plane;
						CV_MAT_ELEM(*Ix1,int,r,c) = max_plane;
#if 1
						// 2x2 advance window, loop-unrolled
						bool gb1, gb2;
						if ( ( gb1 = r < ZHat->rows-1 ) ) {
							if ( CV_MAT_ELEM(*Ix1,int,r+1,c) == 0 ) {
								CV_MAT_ELEM(*Ix1,int,r+1,c) = max_plane;
							}
						}
						if ( ( gb2 = c < ZHat->cols-1 ) ) {
							if ( CV_MAT_ELEM(*Ix1,int,r,c+1) == 0 ) {
								CV_MAT_ELEM(*Ix1,int,r,c+1) = max_plane;
							}
						}
						if ( gb1 && gb2 ) {
							if ( CV_MAT_ELEM(*Ix1,int,r+1,c+1) == 0 ) {
								CV_MAT_ELEM(*Ix1,int,r+1,c+1) = max_plane;
							}
						}
#else
						// looped advance window
						for ( int ro=r; ro<=min(ZHat->rows-1,r+1); ++ro ) {
							for ( int co=c; co<=min(ZHat->cols-1,c+1); ++co ) {
								if ( CV_MAT_ELEM(*Ix1,int,ro,co) == 0 ) {
									CV_MAT_ELEM(*Ix1,int,ro,co) = 1 + max_it->first;
								}
							}
						}
#endif
					}
				}
			}
		}
	}
#if CZH_DEBUG_OUT
	fo = fopen("yyy_Ix1_out.out","wb");
	for ( int c=0; c<Ix1->cols; ++c ) {
		for ( int r=0; r<Ix1->rows; ++r ) {
			foi = CV_MAT_ELEM(*Ix1,int,r,c);
			fwrite(&foi,sizeof(int),1,fo);
		}
	}
	fclose(fo);
#endif
	cvReleaseMat(&Ix1);

#if CZH_DEBUG_OUT
	// output mat to file for sight debugging
	fo = fopen("yyy_Gt_out.out","wb");
	for ( int c=0; c<Gt->cols; ++c ) {
		for ( int r=0; r<Gt->rows; ++r ) {
			fof = CV_MAT_ELEM(*Gt,float,r,c);
			fwrite(&fof,sizeof(float),1,fo);
		}
	}
	fclose(fo);
#endif
	cvReleaseMat(&Gt);

#if CZH_DEBUG_OUT
	// output mat to file for sight debugging
	fo = fopen("yyy_ZHat_out.out","wb");
	for ( int c=0; c<ZHat->cols; ++c ) {
		for ( int r=0; r<ZHat->rows; ++r ) {
			fof = CV_MAT_ELEM(*ZHat,float,r,c);
			fwrite(&fof,sizeof(float),1,fo);
		}
	}
	fclose(fo);
#endif
}

vector<vector<double> > svlPlaneClustering::getRectangularFeatures(vector<CvRect> &rects) {
	vector<vector<double> > ret;
	ret.reserve(rects.size());
	for ( vector<CvRect>::const_iterator i=rects.begin(); i!=rects.end(); ++i ) {
		ret.push_back( getRectangularFeatures(i->x, i->y, i->width, i->height) );
	}
	return ret;
}

vector<double> svlPlaneClustering::getRectangularFeatures(int x, int y, int w, int h, FILE *fo) {
	// calculate 3D support features for the rectangle region specified
	
	double min_z = 1e12, max_z = -1e12;
	int xx, yy;
	// copy out depth values for in situ processing
	vector<double> deps;
	deps.reserve(w*h);
	for ( xx=x; xx<x+w; ++xx ) {
		for ( yy=y; yy<y+h; ++yy ) {
			deps.push_back((double)CV_MAT_ELEM(*_Z,float,yy,xx));
		}
	}
	if ( fo ) {
		fwrite(&x,sizeof(int),1,fo);
		fwrite(&y,sizeof(int),1,fo);
		fwrite(&w,sizeof(int),1,fo);
		fwrite(&h,sizeof(int),1,fo);
		for ( int i=0; i<(int)deps.size(); ++i ) {
			fwrite(&deps[i],sizeof(double),1,fo);
		}
	}

	vector<double> deps_copy(deps);
	// find interquartile distance, hist bounds, bin size
	int gi = (int)deps.size()-1, q1, q3;
	double dq1, dq3, inc_min, inc_max, gd;
	q1 = _sva.select_ordered<double>(deps_copy, 0, gi, (gi>>2)); // first quartile
	dq1 = deps_copy[q1];
	q3 = _sva.select_ordered<double>(deps_copy, 0, gi, (gi>>1)+(gi>>2)); // third quartile
	dq3 = deps_copy[q3];
	gd = dq3-dq1; // histogram halfspan
	inc_min = dq1 - gd/2.0;
	inc_max = dq3 + gd/2.0;

	// get stats on the passed region
	svlPoint3d p_mean(0.0), n_mean(0.0), n_var(0.0);
	vector<svlPoint3d> P, N;
	vector<int> ix; // vector of passed indices
	ix.reserve(w*h);
	P.reserve(w*h);
	N.reserve(w*h); // max number of things
	int ii = 0, ix1;
	bool gb;
	for ( xx=x; xx<x+w; ++xx ) {
		for ( yy=y; yy<y+h; ++yy ) {
			if ( (gb = deps[ii] >= inc_min && deps[ii] <= inc_max) ) { // accumulate this point, support and stuff ...
				if ( (ix1=CV_MAT_ELEM(*_Ix,int,yy,xx)) == -1 )
					continue;
				ix.push_back(ix1);
				P.push_back(svlPoint3d(CV_MAT_ELEM(*_X,float,yy,xx),
									   CV_MAT_ELEM(*_Y,float,yy,xx),
									   CV_MAT_ELEM(*_Z,float,yy,xx)));
				p_mean += P.back();
				//p_mean += (*_reconstP)[ix1];
				N.push_back(svlPoint3d(CV_MAT_ELEM(*_nX,float,yy,xx),
									   CV_MAT_ELEM(*_nY,float,yy,xx),
									   CV_MAT_ELEM(*_nZ,float,yy,xx)));
				N.back().abs();
				n_mean += N.back();
				//n_mean += (*_reconstN)[ix1];
				n_var += N.back() * N.back();
			}
			if ( fo ) fwrite(&gb,sizeof(bool),1,fo); // write out point included or not
			++ii;
		}
	}

#define N_FEAT 14
	vector<double> features(N_FEAT,-1.0);
	if ( (int)ix.size() == 0 )
		return features;

	p_mean /= (double)ix.size();
	n_mean /= (double)ix.size();
	n_var = n_var/(double)ix.size() - n_mean*n_mean; // mean and variance of normal vector
	n_var.ptwise_sqrt(); // square root to get std dev
	n_mean.normalize();

	// project mean position vector onto close planes:
	double minth = 1e6, mintv = 1e6, gd1;
	int n_feat = 3, minhi = -1, minvi = -1;
	
	features[0] = n_mean[0]; // mean normal vector
	features[1] = n_mean[1];
	features[2] = n_mean[2];
	features[3] = n_var[0]; // normal std dev
	features[4] = n_var[1];
	features[5] = n_var[2];
	//vector<double> perp_dists( n_feat * _planes.size(),1e6);
	// project onto closest horizontal/vertical planes

	for ( int i=0; i<(int)_camPlanes.size(); ++i ) {
		gd = 0.0;
		//fwrite(&i,sizeof(int),1,fut);
		if ( _camPlanes[i].isHorizontal() ) {
			if ( fo ) fwrite(&(gi=4),sizeof(int),1,fo);
			if ( (gd=_camPlanes[i].percentSupport(P)) >= 0.5 ) { // must have at least 50% overlap to be compared
				if ( (gd1=_camPlanes[i].minAbsPerpDist(P)) < minth ) {
					minth = gd1;
					minhi = i;
					features[6] = gd;
					features[7] = gd1;
				}
			}
		} else if ( _camPlanes[i].isVertical() ) {
			if ( fo ) fwrite(&(gi=1),sizeof(int),1,fo);
			if ( (gd=_camPlanes[i].percentSupport(P)) >= 0.5 ) {// minth ) {
				if ( (gd1=_camPlanes[i].minAbsPerpDist(P)) < mintv ) {
					mintv = gd1;
					minvi = i;
					features[10] = gd;
					features[11] = gd1;
				}
			}
		} else {
			if ( fo ) fwrite(&(gi=0),sizeof(int),1,fo);
		}
		fwrite(&(gd=_camPlanes[i].minAbsPerpDist(P)),sizeof(double),1,fo);
	}

	// for nearest h/v planes, calc height mean and std dev
	gd = gd1 = 0.0;
	if ( minhi != -1 )
		_camPlanes[minhi].heightMeanVar(P, features[8], features[9]);
	if ( minvi != -1 )
		_camPlanes[minvi].heightMeanVar(P, features[12], features[13]);

	if ( fo ) { // write features out to ... thing
		int gi;
		fwrite(&inc_min,sizeof(double),1,fo);
		fwrite(&inc_max,sizeof(double),1,fo);
		fwrite(&minvi,sizeof(int),1,fo);
		fwrite(&minhi,sizeof(int),1,fo);
		fwrite(&(gi=(int)features.size()),sizeof(int),1,fo);
		for ( int i=0; i<(int)features.size(); ++i ) {
			fwrite(&features[i],sizeof(double),1,fo);
		}	
	}
	return features;
}

