/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlPatchUtils.cpp
** AUTHOR(S):   Ian Goodfellow <ia3n@cs.stanford.edu>
**              Olga Russakovsky <olga@cs.stanford.edu>
**      based on applications by Stephen Gould and Sid Batra
**             
** DESCRIPTION:
**   Commonly re-used subroutines for processing command line arguments related
**   to patch types, mapping from channel types to patch types, or reading
**   and writing image files in the patch based object detector pipeline
**
*****************************************************************************/
#include "svlPatchUtils.h"


svlPatchUtil::svlPatchUtil(unsigned flags) : 
  _baseExt(".jpg"), //_baseExt("")
  _cliBaseExt(false),
  _defaultEdge(false),
  _depthMapWidth(640),
  _depthMapHeight(480),
  _width(-1),
  _height(-1),
    _depthMapExt(".reconstructed.z.txt"),
    _imageExt(".jpg"),
  _cliDepthChannel(false),
    _hasDepthChannel(false),
    _usedImageExt(false),
  _usedDepthMapExt(false),
  _usedDeprecatedCh(false),
  _usedNewCh(false)
{

  //Load the mapping from command line text to channel types
    _textToChannelType["INTENSITY"] = SVL_CHANNEL_INTENSITY;
    _textToChannelType["EDGE"] = SVL_CHANNEL_EDGE;
    _textToChannelType["DEPTH"] = SVL_CHANNEL_DEPTH;
    _textToChannelType["ABS"] = SVL_CHANNEL_ABS;
    _textToChannelType["ABSG"] = SVL_CHANNEL_ABSG;
    _textToChannelType["ABS_HSV"] = SVL_CHANNEL_ABS_HSV;
    _textToChannelType["ABS_YCrCb"] = SVL_CHANNEL_ABS_YCrCb;
    _textToChannelType["ABS_BG"] = SVL_CHANNEL_ABS_BG;
    _textToChannelType["ABS_BR"] = SVL_CHANNEL_ABS_BR;
    _textToChannelType["ABS_GR"] = SVL_CHANNEL_ABS_GR;
    _textToChannelType["ABS_HS"] = SVL_CHANNEL_ABS_HS;
    _textToChannelType["ABS_HV"] = SVL_CHANNEL_ABS_HV;
    _textToChannelType["COLOR"] = SVL_CHANNEL_INTENSITY_H;
  

    //Load the mapping from channel type to patch type
    _channelTypeToPatchType[SVL_CHANNEL_INTENSITY] = SVL_INTENSITY_PATCH;
    _channelTypeToPatchType[SVL_CHANNEL_EDGE] = SVL_INTENSITY_PATCH;
    _channelTypeToPatchType[SVL_CHANNEL_DEPTH] = SVL_DEPTH_PATCH;
    _channelTypeToPatchType[SVL_CHANNEL_ABS] = SVL_ABS_PATCH;
    _channelTypeToPatchType[SVL_CHANNEL_ABSG] = SVL_ABS_PATCH;
    _channelTypeToPatchType[SVL_CHANNEL_ABS_HSV] = SVL_ABS_PATCH;
    _channelTypeToPatchType[SVL_CHANNEL_ABS_YCrCb] = SVL_ABS_PATCH;
    _channelTypeToPatchType[SVL_CHANNEL_ABS_BG] = SVL_ABS_PATCH;
    _channelTypeToPatchType[SVL_CHANNEL_ABS_BR] = SVL_ABS_PATCH;
    _channelTypeToPatchType[SVL_CHANNEL_ABS_GR] = SVL_ABS_PATCH;
    _channelTypeToPatchType[SVL_CHANNEL_ABS_HS] = SVL_ABS_PATCH;
    _channelTypeToPatchType[SVL_CHANNEL_ABS_HV] = SVL_ABS_PATCH;
    _channelTypeToPatchType[SVL_CHANNEL_INTENSITY_H] = SVL_INTENSITY_PATCH;

    if (flags & SVL_PATCH_DEFAULT_EDGE)
	_defaultEdge = true;
     

    if (flags & SVL_PATCH_CLI_DEPTHCHANNEL)
      _cliDepthChannel = true;

    if (flags & SVL_PATCH_CLI_BASE_EXT)
      _cliBaseExt = true;
}


bool svlPatchUtil::parseCommandLineItem(int & argc, char ** & args)
{
  if (_cliBaseExt && !strcmp(*args, "-baseExt"))
    {
      _baseExt = * (++args);
      argc--;
    }
  else if (!strcmp(*args, "-imageExt"))
    {
      SVL_LOG(SVL_LOG_WARNING, "-imageExt is deprecated, use -ch INTENSITY <ext> instead");
      _usedImageExt = true;
      _imageExt = string(*(++args)); argc--;
    } 
  else if (!strcmp(*args, "-depthMapExt"))
    {
      SVL_LOG(SVL_LOG_WARNING, "-depthMapExt is deprecated, use -ch DEPTH <ext> instead");
      _depthMapExt = *(++args); 
      argc--;
      _usedDepthMapExt = true;
    }
  else if (!strcmp(*args, "-dmsize")) 
    {
      _depthMapWidth = atoi(*(++args)); argc--;
      _depthMapHeight = atoi(*(++args)); argc--;
    }
  else if (_cliDepthChannel && !strcmp(*args, "-depthChannel"))
    {
      SVL_LOG(SVL_LOG_WARNING, "-depthChannel is deprecated. use -ch DEPTH <ext> instead");
      _hasDepthChannel = true;
    }
  else if (!strcmp(*args, "-ch"))
      {
	//Advance to the first argument to the flag-- should be a type if using the new system or a number if using the deprecated system
	++args;
	argc--;

	if (argc == 0)
	  SVL_LOG(SVL_LOG_FATAL, "No arguments to -ch flag");

	//See if we can parse the next argument as a type
	if (_textToChannelType.find(*args) != _textToChannelType.end())
	  {
	    //We parsed the argument as a type, so use the new system

	    //Record that we used this system
	    _usedNewCh = true;

	    //Since we're using the new system, make sure the deprecated system hasn't been used
	    if (_usedDeprecatedCh)
	      SVL_LOG(SVL_LOG_FATAL, "Attempted to use deprecated -ch format with other non-deprecated channel arguments. This mixture of syntax is not supported");

	    //Get the channel type
	    svlChannelType cType = _textToChannelType[*args];

	    //Read to the next argument
	    args++;
	    argc--;

	    if (argc == 0)
	      SVL_LOG(SVL_LOG_FATAL, "Only one argument to -ch flag, two required");

	    //Save the type
	    _channelTypes.push_back(cType);
	    
	    //Save the patch type
	    _channelPatchTypes.push_back( _channelTypeToPatchType[cType] );

	    //Save the extension for this channel (the next argument)
	    _channelExts.push_back(*args);
	  }
	else
	  {
	    //In this case, we could not parse the next argument as a type
	    //This could be a syntax error or the deprecated usage of -ch
	    
	    int numChannels = atoi(*args);
	    
	    //Check to see if this is a syntax error
	    if (numChannels == 0)
	      SVL_LOG(SVL_LOG_FATAL, "Error while processing -ch flag. Either first arg (\""<<(*args)<<"\")is syntax error, or request for 0 channels in deprecated usage of -ch");

	    //If we got to here, this is the deprecated usage

	    SVL_LOG(SVL_LOG_WARNING, "using deprecated version of -ch");

	    //Record that this deprecated argument has been used
	    _usedDeprecatedCh = true;
	 
	    //Make sure no other channels have been created by other means
	    if (_channelTypes.size() != 0)
	      SVL_LOG(SVL_LOG_FATAL, "Attempted to use deprecated -ch format with other non-deprecated channel arguments. This mixture of syntax is not supported");
	    

	    //Read all of the channel types
	    for (int i = 0; i < numChannels; i++)
	      {
		args++;
		argc--;

		//Check that we haven't read past the end of input
		if (argc == 0)
		  SVL_LOG(SVL_LOG_FATAL, "" << numChannels << " channels requested with deprecated version of -ch, but only "<<i<<" channel types specified");

		//Check that the argument is a valid type
		if (_textToChannelType.find(*args) == _textToChannelType.end())
		  SVL_LOG(SVL_LOG_FATAL, "Invalid channel type \""<<*args<<"\" requested in call to deprecated version of -ch")
		else
		  {
		    //Get the channel type
		    svlChannelType cType = _textToChannelType[*args];
		    //Get the patch type
		    svlPatchDefinitionType pType = _channelTypeToPatchType[cType];

		    //Store the channel type and patch type (channel extension will have to be back-patched later because this is the deprecated syntax)
		    _channelTypes.push_back(cType);
		    _channelPatchTypes.push_back(pType);
		  }
	      }//closes loop over all the channel type arguments
	  } //closes processing of deprecated version of ch arg or syntax error following -ch arg
      } //closes processing of -ch arg
  else
    return false;

  return true;
}

void svlPatchUtil::finishCommandLine()
{
  if (_usedNewCh)
    {
      assert(!_usedDeprecatedCh); //We already have checks against this so if both are true something is wrong with the parser
     

      //We should check the usage though
      if (_usedImageExt)
	SVL_LOG(SVL_LOG_FATAL, "Attempted to use new -ch argument with deprecated -imageExt argument. This hybrid usage is not supported");
      if (_usedDepthMapExt)
	SVL_LOG(SVL_LOG_FATAL, "Attempted to use new -ch argument with deprecated -depthMapExt argument. This hybrid usage is not supported");
      if (_hasDepthChannel)
	SVL_LOG(SVL_LOG_FATAL, "Attempted to use new -ch argument with deprecated -depthChannel argument. This hybrid usage is not supported.");
    }

  if (_channelTypes.size())
    {
      //The channels have been populated already, using some form of -ch

      //If the user used the deprecated version of -ch, then the extensions won't be populated yet
      if (!_channelExts.size())
	{
	  //If this happens it should only be because of use of the deprecated version of -ch
	  assert(_usedDeprecatedCh);

	  //Fill in the extension based on the type for all channels
	  for (unsigned i = 0; i < _channelTypes.size(); i++)
	    {
	      switch (_channelTypes[i])
		{
		case SVL_CHANNEL_DEPTH:
		  _channelExts.push_back(_depthMapExt);
		  break;
		default:
		  _channelExts.push_back(_imageExt);
		  break;
		} //closes the switch over channel types
	    }//closes the loop over channels
	}//closes handling of unpopulated channelExts
      else
	assert( !_usedDeprecatedCh); //If the extensions were populated, then we must not have used the deprecated -ch
    }//closes case where channels are already populated
  else
    {
      //The channels have not been populated yet, so we must fill in the defaults

      //Always assume there is an INTENSITY channel.
      _channelTypes.push_back(SVL_CHANNEL_INTENSITY);
      _channelPatchTypes.push_back(SVL_INTENSITY_PATCH);
      //Default its extension to baseExt, but for compatibility with old scripts, respect the deprecated imageExt flag if it was called
      if (_usedImageExt)
	_channelExts.push_back(_imageExt);
      else
	_channelExts.push_back(_baseExt);

      //If configured to assume there is an edge patch, add an edge patch
      if (_defaultEdge)
	{
	  _channelTypes.push_back(SVL_CHANNEL_EDGE);
	  _channelPatchTypes.push_back(SVL_INTENSITY_PATCH);
	  if (_usedImageExt)
	    _channelExts.push_back(_imageExt);
	  else
	    _channelExts.push_back(_baseExt);
	}

      //For compatibility with old scripts, respect the deprecated depthChannel flag if it was called
      if (_hasDepthChannel)
	{
	  _channelTypes.push_back(SVL_CHANNEL_DEPTH);
	  _channelPatchTypes.push_back(SVL_DEPTH_PATCH);
	  _channelExts.push_back(_depthMapExt);
	}
    }

  //Patch any uses of "-" with baseExt
  for (unsigned i = 0; i < _channelExts.size(); i++) {
    if (_channelExts[i] == "-")
      _channelExts[i] = _baseExt;
  }
  

  //At this point all of our channel definition vectors should be complete
  //Check that their sizes match
  assert(_channelTypes.size() == _channelPatchTypes.size());
  assert(_channelPatchTypes.size() == _channelExts.size());
  //Check that they are non-empty
  assert(_channelTypes.size() != 0);


  //Prepare for loading files later on
  for (unsigned i = 0; i < _channelExts.size(); i++)
    {
      _files.push_back(vector<string> ());
      _eTree.addChannel(i, _channelExts[i]);
    }

}

void svlPatchUtil::readDir(const char * dir)
{
  _dir = dir;

  _eTree.readDirectory(dir);

  for (unsigned i = 0; i < _channelExts.size(); i++)
    {
      _files[i].clear();
      _eTree.getFilesForChannel(i, _files[i]);
    }


  //Check that the files all correspond to each other
  if (_files.size() > 1)
    {
      for (unsigned i = 1; i < _files.size(); i++)
	{
	  assert(_files[i].size() == _files[0].size());
	  
	  for (unsigned j = 0; j < _files[0].size(); j++)
	    {
	      assert(!strcmp(stripExtension(_files[0][j], _channelExts[0]).c_str(), stripExtension(_files[i][j], _channelExts[i]).c_str()));
	    }
	}
    }
}

void svlPatchUtil::getAllFrames(vector<vector<IplImage *> > & output) const
{
  output.clear();

  bool newFrameNeeded = true;
  for (unsigned int i = 0; i < _files[0].size(); i++)
    {
      if (newFrameNeeded)
	output.push_back(vector<IplImage *>() );

      newFrameNeeded = getFrame(i, output.back());
    }
}

bool svlPatchUtil::getFrame(unsigned idx, vector<IplImage *> & output) const
{
  SVL_LOG(SVL_LOG_MESSAGE, "Getting frame " << stripExtension(_files[0][idx], _channelExts[0]));

  vector<string> filenames;

  for (unsigned ch = 0; ch < _channelTypes.size(); ch++)
    filenames.push_back(  _dir + "/" + _files[ch][idx] );

  return getFrameFromFiles(filenames, output);
}

bool svlPatchUtil::getFrameFromBasename(const string & baseName, vector<IplImage *> & output) const
{
  string strippedName = baseName.substr(0,baseName.size()-_baseExt.size());

  vector<string> filenames;

  for (unsigned ch = 0; ch < _channelExts.size(); ch++)
    filenames.push_back(strippedName + _channelExts[ch]);

  return getFrameFromFiles(filenames, output);
}

bool svlPatchUtil::getFrameFromFiles(const vector<string> & files, vector<IplImage *> & output) const
{

  if (files.size() != _channelTypes.size())
    return false;

  for (unsigned ch = 0; ch < _channelTypes.size(); ch++)
    {
      //TODO-- this implementation doesn't attempt to avoid loading the same file multiple times. Could restore that feature
      //note-- reloading the same file doesn't seem to incur much of a performance penalty in my experiments, in fact,
      //my new code came out faster than the original in all three trials that I ran. I assume this is because the file is reloaded soon after
      //the initial load and is still in the disk buffer. The short story is that I don't think avoiding reloading the same file
      //is a high priority. -IG

      // Certainly makes the code a heck of a lot cleaner and happier. -- OR

      const string & filename = files[ch];

      // Intensity patches are loaded in grayscale, abs patches are loaded in color.
      // The two exceptions are ABSG, which are gradient-based but should be grayscale,
      // and INTENSITY_H, which are like intensity patches except using hue, not intensity,
      // and thus must be loaded in color first.
      svlPatchDefinitionType type = _channelPatchTypes[ch];
      if (_channelTypes[ch] == SVL_CHANNEL_ABSG)
	type = SVL_INTENSITY_PATCH;
      else if (_channelTypes[ch] == SVL_CHANNEL_INTENSITY_H)
	type = SVL_ABS_PATCH;

      IplImage * img = loadPatch(filename, type);

      if (img == NULL)
	{
	  SVL_LOG(SVL_LOG_MESSAGE, "...failed to load " << filename);

	  for (unsigned ch2 = 0; ch2 < ch; ch2++)
	    cvReleaseImage(&output[ch2]);

	  output.clear();

	  return false;
	}
      else
	{
	  if ((img->width != (int) _width) || (img->height != (int) _height))
	    {
	      if (_height != -1)
		resizeInPlace(&img, _height, _width);
	    }

	  if (_channelTypes[ch] == SVL_CHANNEL_EDGE) {
	    IplImage *temp = cvCloneImage(_softEdgeCalculator.processImage(img));
	    cvReleaseImage(&img);
	    img = temp;
	  }
	  
	  // if was attempted to be loaded in color, do whatever is necessary
	  if (type == SVL_ABS_PATCH) {
	    _converter.changeColorModel(img, _channelTypes[ch]);
	  }

	  output.push_back(img);	  
	}    
    }


  return true;
}


IplImage * svlPatchUtil::loadPatch(const string & path, svlPatchDefinitionType type) const
{
  switch (type)
    {
    case SVL_INTENSITY_PATCH:
      return cvLoadImage(path.c_str(), CV_LOAD_IMAGE_GRAYSCALE);
    case SVL_DEPTH_PATCH:
      return readMatrixAsIplImage(path.c_str(), _depthMapWidth, _depthMapHeight);
    case SVL_ABS_PATCH:
      return cvLoadImage(path.c_str(), CV_LOAD_IMAGE_UNCHANGED);
    default:
      assert(false); //should be unreached
    }
  assert(false); //should be unreached
  return NULL; //just included to suppress compiler warnings
}

void svlPatchUtil::setDepthMapSize(int width, int height)
{
  _depthMapWidth = width;
  _depthMapHeight = height;
}

void svlPatchUtil::setResizeSize(int width, int height)
{
  _width = width;
  _height = height;

  if (_depthMapWidth == -1)
    setDepthMapSize(width, height);
}

unsigned svlPatchUtil::numFrames() const
{
  if (_files.size())
    return (unsigned)_files[0].size();
  return 0;
}


string svlPatchUtil::stripExtension(const string & str, const string & ext)
{
  return str.substr(0, str.size()-ext.size());
}








































	


void svlColorConverter::changeColorModel(IplImage *&image, svlChannelType cs) const {
  if (image->nChannels == 1) return; // nothing more can be done, just leave as grayscale

  // invariant: the current color space is BGR

  IplImage *result;
  switch(cs) {
    // loaded correctly, all is good
  case SVL_CHANNEL_ABS:
    return;

    // conversions using openCV
  case SVL_CHANNEL_ABS_HSV:
    cvCvtColor(image, image, CV_BGR2HSV);
    return;
  case SVL_CHANNEL_ABS_YCrCb:
    cvCvtColor(image, image, CV_BGR2YCrCb);
    return;
    
    // conversion to 2 colors from BGR by averaging
  case SVL_CHANNEL_ABS_BG:
    averageColorChannels(image, 3);
    return;
  case SVL_CHANNEL_ABS_BR:
    averageColorChannels(image, 2);
    return;
  case SVL_CHANNEL_ABS_GR:
    averageColorChannels(image, 1);
    return;
  
    // conversion to 2 colors from HSV by deleting a channel
  case SVL_CHANNEL_ABS_HS:
    cvCvtColor(image, image, CV_BGR2HSV);
    deleteChannel(image, 3);
    return;
  case SVL_CHANNEL_ABS_HV:
    cvCvtColor(image, image, CV_BGR2HSV);
    deleteChannel(image, 2);
    return;

    // conversion to intensity-type patch from HSV by leaving only the hue
  case SVL_CHANNEL_INTENSITY_H:
    cvCvtColor(image, image, CV_BGR2HSV);
    leaveSingleChannel(image, 1);
    return;

    // for completeness, this should in practice never happen
  case SVL_CHANNEL_INTENSITY:
  case SVL_CHANNEL_ABSG:
    result = cvCreateImage(cvSize(image->width, image->height), image->depth, 1);
    cvCvtColor(image, result, CV_BGR2GRAY);
    cvReleaseImage(&image);
    image = result;
    return;
    
  case SVL_CHANNEL_EDGE:
  case SVL_CHANNEL_DEPTH:
    SVL_ASSERT_MSG(false, "Can't perform this conversion; unclear why it's requested in the first place");
  default: SVL_ASSERT(false); // to suppress warnings
  }
}


void svlColorConverter::deleteChannel(IplImage *&img, int i) const {
  assert(img->nChannels == 3 && i >= 1 && i <= 3);
  IplImage *result = cvCreateImage(cvSize(img->width, img->height), img->depth, 2);
  
  int c1 = (i == 1) ? 2 : 1;
  int c2 = (i == 3) ? 2 : 3;
  
  IplImage *tmp1 = cvCreateImage(cvSize(img->width, img->height), img->depth, 1);
  cvSetImageCOI(img, c1);
  cvCopy(img, tmp1);
  
  IplImage *tmp2 = cvCreateImage(cvSize(img->width, img->height), img->depth, 1);
  cvSetImageCOI(img, c2);
  cvCopy(img, tmp2);

  cvMerge(tmp1, tmp2, NULL, NULL, result);
  
  cvReleaseImage(&img);
  cvReleaseImage(&tmp1);
  cvReleaseImage(&tmp2);
  img = result;
}

void svlColorConverter::leaveSingleChannel(IplImage *&img, int i) const {
  SVL_ASSERT(img->nChannels == 3 && i >= 1 && i <= 3);
  IplImage *result = cvCreateImage(cvSize(img->width, img->height), img->depth, 1);
  
  cvSetImageCOI(img, i);
  cvCopy(img, result);
  
  cvReleaseImage(&img);
  img = result;
}

// The image will now have two color channels: the first one i,
// the second one an average of the other two channels in the original
void svlColorConverter::averageColorChannels(IplImage *&img, int i) const {
  SVL_ASSERT(img->nChannels == 3 && i >= 1 && i <= 3 && img->depth == IPL_DEPTH_8U);
  IplImage *result = cvCreateImage(cvSize(img->width, img->height), img->depth, 2);
  
  // Do floats if depth is not 8U
  uchar* data;
  int step;
  CvSize size;
  cvGetRawData( img, &data, &step, &size);
  step /= sizeof(data[0]);
  
  uchar *newdata;
  int newstep;
  cvGetRawData(result, &newdata, &newstep);
  newstep /= sizeof(newdata[0]);
  
  i = i-1;
  int c1 = (i+1)%3;
  int c2 = (i+2)%3;
  float w1 = bgr_weights[c1];
  float w2 = bgr_weights[c2];
  float total = w1 + w2;
  
  for(int y = 0; y < size.height; y++, data += step, newdata += newstep ) {
    for(int x = 0; x < size.width; x++) {
      newdata[2*x] = data[3*x+i];
      newdata[2*x + 1] = (uchar)((w1 * data[3*x + c1] + w2 * data[3*x + c2]) / total);
    }
  }
  
  cvReleaseImage(&img);
  img = result;
}
